#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import time
import logging
import datetime
import threading

from libraries.enum.auraenumerations import ScheduleEntryType


class LiquidSoapInitThread(threading.Thread):
    logger = None
    active_entry = None
    liquidsoapcommunicator = None

    # ------------------------------------------------------------------------------------------ #
    def __init__(self, liquidsoapcommunicator, active_entry):
        threading.Thread.__init__(self)
        self.logger = logging.getLogger("AuraEngine")
        self.liquidsoapcommunicator = liquidsoapcommunicator
        self.active_entry = active_entry

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        try:
            # sleep needed, because the socket is created too slow by liquidsoap
            time.sleep(1)
            self.logger.info("Waited 1s for liquidsoap. Jez soit a si gspian")

            # enable lqs transaction
            self.liquidsoapcommunicator.enable_transaction()

            # wait another second. lqs really starts slow.. be prepared you liquidsoap you!
            time.sleep(1)

            # set some parameters
            self.set_start_parameters()

            # set active
            self.set_active_show()

            # disable lqs transaction again
            self.liquidsoapcommunicator.disable_transaction()

            # the rest of the system now can use liquidsoap connection
            self.liquidsoapcommunicator.is_liquidsoap_running = True
        except Exception as e:
            self.logger.critical("Liquidsoap connection ERROR! Restart LQ Server! Reason: "+str(e))

    def set_active_show(self):
        if self.active_entry is not None:
            self.logger.info("LiquidSoapInitThread sets activechannel: " + str(self.active_entry))

            channel = self.active_entry.type

            # have to seek?
            if channel == ScheduleEntryType.FILESYSTEM:
                # calc how many seconds were missed
                now_unix = time.mktime(datetime.datetime.now().timetuple())
                seconds_to_seek = now_unix - self.active_entry.start_unix

                # and seek these seconds forward
                if seconds_to_seek > 0:
                    self.liquidsoapcommunicator.playlist_seek(seconds_to_seek)

            # finally make something hearable :-)
            if channel != "" and channel is not None:
                # activate http stream if needed
                self.liquidsoapcommunicator.http_start_stop(channel == ScheduleEntryType.STREAM)
                # finally set the volume up
                self.liquidsoapcommunicator.channel_volume(channel.value, self.active_entry.volume)
            else:
                self.logger.error("Channel is NULL or empty! Cannot set ")

        else:
            self.logger.warning("No active entry in the scheduler! Is a programme loaded?")

    def set_start_parameters(self):
        # reset channels and reload them
        channels = self.liquidsoapcommunicator.reload_channels()

        # for all available channels
        for c in channels:
            # set volume to zero
            self.liquidsoapcommunicator.channel_volume(c, "0")
            # and activate this channel
            self.liquidsoapcommunicator.channel_activate(c, True)

        # setting init params like a blank file..
        install_dir = self.liquidsoapcommunicator.config.get("install_dir")
        self.liquidsoapcommunicator.playlist_push(install_dir + "/configuration/blank.flac")
        # .. or the radio fro stream (it is overwritten as soon as one http overtake is planned)
        #self.liquidsoapcommunicator.set_http_url("http://stream.fro.at/fro-128.ogg")
